# Copyright 2004 Tom Rothamel <tom-renpy@onegeek.org>
# 
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# s of the game script.
import re

def nextify_list(l, next):
    """
    This iterates through each pair of elements l[n] and l[n+1] in
    a list of fragments, calling l[n].set_next(l[n+1]). Finally,
    this calls l[-1].set_next(next).
    
    It should be called by the set_next method of any fragment
    class that transfers control to a list.
    """
    
    for a, b in zip(l, l[1:]):
        a.set_next(b)

    l[-1].set_next(next)
    

# A map from fragment name to the fragment instance.
frag_dict = {}

# A list of all fragments known to the system.
frag_list = []

def make_frag_dict():
    """
    Builds up a dictionary mapping fragment name to fragment object.
    """

    for f in frag_list:
        frag_dict[f.name] = f

def lookup(name):
    """
    Returns the fragment corresponding to name.
    """

    return frag_dict[name]
    
class Fragment(object):
    """
    A fragment represents a portion of the script.

    @ivar next: The fragment that will be executed after this one is
    complete.

    @ivar name: The name of this fragment. This may be changed by the
    label command.
    """

    frag_serial = 0

    def __init__(self):
        self.name = "frag_%04d" % Fragment.frag_serial
        Fragment.frag_serial += 1
        
        self.next = None
        frag_list.append(self)

    def set_next(self, next):
        self.next = next

    def __repr__(self):
        return "<%s '%s'>" % (self.__class__.__name__, self.name)

    def run(self, e):
        print "s of type %s not implemented." % self.__class__.__name__
        return self.next

    def check_jump(self):
        pass

# Control fragments.

class Python(Fragment):
    """
    This is a fragment that executes a block of python code.
    """
    
    def __init__(self, code):
        super(Python, self).__init__()        

        # We need to fix up the indentation of the python code.
        lines = [ ]
        indent = ""

        if code.startswith("\020"):
            code = code[1:-1]

        for l in code.split('\n'):
            while l.startswith("\020"):
                indent = indent + "    "
                l = l[1:]

            while l.startswith("\021"):
                indent = indent[:-4]
                l = l[1:]
            
            lines.append(indent + l)

        self.code = "\n".join(lines)

    def run(self, e):
        e.do_python(self.code)
        return self.next

class Pass(Fragment):

    def run(self, e):
        return self.next

class Label(Fragment):

    def __init__(self, name):
        super(Label, self).__init__()
        self.name = name

    def run(self, e):
        return self.next

class Jump(Fragment):
    """
    A fragment that jumps to a target when executed, used to transfer
    control from one location to another in the program.
    """
    
    def __init__(self, next_name):
        super(Jump, self).__init__()
        self.next_name = next_name

    def set_next(self, next):
        pass

    def run(self, e):
        return lookup(self.next_name)

    def check_jump(self):
        if self.next_name not in frag_dict:
            print "Jump target '%s' does not exist." % self.next_name

class If(Fragment):
    """
    A fragment that evaluates a python expression, and then jumps to
    either a true or false branch, depending on the result.
    """

    def __init__(self, cond, true, false):
        super(If, self).__init__()

        self.cond = cond
        self.true = true
        self.false = false

    def set_next(self, next):
        nextify_list(self.true, next)

        if (self.false):
            nextify_list(self.false, next)
        else:
            self.next = next

    def run(self, e):
        if e.eval(self.cond):
            return self.true[0]

        if self.false:
            return self.false[0]
        else:
            return self.next

class Call(Fragment):
    def __init__(self, target):
        super(Call, self).__init__()
        self.target = target

    def run(self, e):
        e.do_save_return(self.next)
        return lookup(self.target)

class Return(Fragment):

    def __init__(self):
        super(Return, self).__init__()

    def run(self, e):
        return None
    
# UI s

class Dialogue(Fragment):
    """
    This contains a fragment of dialog in a script.

    @ivar who: The name of the character that is doing the speaking
    @ivar what: The dialog being spoken.
    """
    
    def __init__(self, who, what, lookup):
        super(Dialogue, self).__init__()

        self.who = who
        self.what = what
        self.lookup = lookup

    def run(self, e):
        e.do_dialogue(self.who, self.what, self.lookup)
        return self.next

class Menu(Fragment):
    """
    This fragment represents a menu that will be shown to the user.
    """

    def __init__(self, name, label, choices, centered, seenset):
        super(Menu, self).__init__()

        if name:
            self.name = name

        if seenset:
            newchoices = [ ]

            for i, (l, cond, block) in enumerate(choices):
                if cond:
                    cond = "(%d not in %s) and (%s)" % (i, seenset, cond)
                else:
                    cond = "%d not in %s" % (i, seenset)

                block = [ Python("%s.append(%d)" % (seenset, i)) ] + block

                newchoices.append((l, cond, block))

            choices = newchoices

        self.label = label
        self.choices = choices
        self.centered = centered

    def set_next(self, next):
        for l, cond, cmds in self.choices:
            nextify_list(cmds, next)

    def run(self, e):
        choices = [ (label, cmds) for label, cond, cmds in self.choices if
                    cond is None or e.eval(cond) ]

        choice = e.do_menu(self.label,
                           [ i[0] for i in choices ],
                           self.centered)

        return choices[choice][1][0]


# Declaration s

class CharacterDeclaration(Fragment):
    """
    This fragment represents the declaration of a character to the
    system.
    """

    def __init__(self, name):
        super(CharacterDeclaration, self).__init__()
        self.character_name = name

    def run(self, e):
        e.do_character(self.character_name)
        return self.next

class ImageDeclaration(Fragment):
    """
    This fragment represents the declaration of an image to the
    system.
    """

    def __init__(self, thing, code):
        super(ImageDeclaration, self).__init__()
        
        self.thing = thing
        self.code = code

    def run(self, e):
        e.ui_image(self.thing, self.code)
        return self.next

# Scene s

class Scene(Fragment):
    """
    This fragment represents the declaration of a new scene. It may
    also optionally contain a tuple describing something that will
    placed in that scene.
    """

    def __init__(self, thing):
        super(Scene, self).__init__()

        self.thing = thing

    def run(self, e):
        e.do_scene()
        if self.thing:
            e.do_show(self.thing, None)

        return self.next
            

class Show(Fragment):
    """
    This fragment adds something to the scene.
    """

    def __init__(self, thing):
        super(Show, self).__init__()
        self.thing = thing

    def run(self, e):
        e.do_show(self.thing, None)

        return self.next
        
class Hide(Fragment):
    """
    This fragment removes something from the scene.
    """

    def __init__(self, thing):
        super(Hide, self).__init__()
        self.thing = thing
        
    def run(self, e):
        e.do_hide(self.thing)

        return self.next
