# Copyright 2004 Tom Rothamel <tom-renpy@onegeek.org>
# 
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import pygame
from pygame.locals import *

import display
import image
from params import config

class Text(display.Showable):

    def __init__(self, width, text):
        """
        This renders text into a box on the screen. The height of the box
        is the minimum height that still fits all the text.
        
        @param width: The width of the text.
        @param text: The text. A list of tuples, with the first element of
        each tuple being a paragraph of text, the second element being the
        color, and the third element being a parameter.
        """
        
        self.width = width
        self.text = text

        self.layout()

    def set_text(self, text):

        if self.text != text:

            self.text = text
            self.layout()

            display.redraw(0)




    def layout(self):
        lines = []
        
        for l, c, p in self.text:
            words = l.split(' ')
            line = ""
            width = 0

            for w in words:
                tw, th = font.size(line + w + " ")

                if tw > self.width:
                    lines.append((line, c, p, width))
                    line = w + " "
                    width = font.size(w)[0]
                else:
                    line += w + " "
                    width = tw

            lines.append((line, c, p, tw))

        self.lines = lines

    def size(self):
        return (self.width,
                (len(self.lines) - 1) * font_linesize + font_height)

    def draw(self, trans):
        y = 0

        for l, c, p, width in self.lines:

            if config.shadow_offset:
                ssurf = font.render(l, True, config.shadow_color)
                trans.blit(ssurf, config.shadow_offset[0], config.shadow_offset[1] + y)

            surf = font.render(l, True, c)
            trans.blit(surf, 0, y)
            y += font_linesize


    def mouse_translate(self, x, y):

        if x < 0 or y < 0:
            return None

        line = y // font_linesize

        if line >= len(self.lines):
            return None

        text, color, result, width = self.lines[line]

        if x < width:
            return result
        else:
            return None

class Box(display.Showable):

    def class_init(self):

        bb = config.box_border
        surface = image.load_image(config.box_image)
        width, height = surface.get_size()

        def subrect(x0, x1, y0, y1):

            if x0 < 0:
                x0 += width

            if x1 <= 0:
                x1 += width

            if y0 < 0:
                y0 += height

            if y1 <= 0:
                y1 += height

            return surface.subsurface((x0, y0, x1 - x0, y1 - y0))

        self.tl = subrect(0, bb, 0, bb)
        self.tc = subrect(bb, -bb, 0, bb)
        self.tr = subrect(-bb, 0, 0, bb)

        self.cl = subrect(0, bb, bb, -bb)
        self.cc = subrect(bb, -bb, bb, -bb)
        self.cr = subrect(-bb, 0, bb, -bb)

        self.bl = subrect(0, bb, -bb, 0)
        self.bc = subrect(bb, -bb, -bb, 0)
        self.br = subrect(-bb, 0, -bb, 0)

    class_init = classmethod(class_init)

    def __init__(self, width, height):
        self.box_width = width
        self.box_height = height

    def size(self):
        return self.box_width, self.box_height

    def draw(self, trans):

        bb = config.box_border
        width = self.box_width
        height = self.box_height

        def drawrect(surf, x0, x1, y0, y1):

            if x0 < 0:
                x0 += width

            if x1 <= 0:
                x1 += width

            if y0 < 0:
                y0 += height

            if y1 <= 0:
                y1 += height

            w = x1 - x0
            h = y1 - y0 

            s = pygame.transform.scale(surf, (w, h))

            trans.blit(s, x0, y0)

        drawrect(self.tl, 0, bb, 0, bb)
        drawrect(self.tc, bb, -bb, 0, bb)
        drawrect(self.tr, -bb, 0, 0, bb)
    
        drawrect(self.cl, 0, bb, bb, -bb)
        drawrect(self.cc, bb, -bb, bb, -bb)
        drawrect(self.cr, -bb, 0, bb, -bb)    

        drawrect(self.bl, 0, bb, -bb, 0)
        drawrect(self.bc, bb, -bb, -bb, 0)
        drawrect(self.br, -bb, 0, -bb, 0)

class TextBox(Text, Box):
    def __init__(self, width, text, minheight=0):
        Text.__init__(self, width - config.box_hpadding * 2, text)
        tw, th = Text.size(self)

        h = th + config.box_vpadding * 2
        if h < minheight:
            h = minheight
        
        Box.__init__(self, width, h)

    def size(self):
        return Box.size(self)

    def draw(self, trans):
        Box.draw(self, trans)
        Text.draw(self, trans.move(config.box_hpadding, config.box_vpadding))

    def mouse_translate(self, x, y):
        return Text.mouse_translate(self, x - config.box_hpadding, y - config.box_vpadding)

def _list_text(s, color, parameter=None):
    """
    This splits the text into a list of paragraphs at newlines. Each
    entry in the list is assigned the specified color and parameter.
    """

    rv = []
    for i in s.split('\n'):
        rv.append((i, color, parameter))

    return rv

class MenuBox(TextBox):

    def __init__(self, label, choices, width=None):

        if width is None:
            width = config.menu_box_width

        self.choice = 0
        self.choices = choices
        self.label = label
    
        TextBox.__init__(self, width, self.menu_layout(),
                         minheight=config.menu_box_minheight)

    def mouse_event(self, button, x, y):
        choice = self.mouse_translate(x, y)

        if choice is None:
            return False

        self.select(choice)
        return button

    def key_event(self, e):

        if e.key == K_UP:
            self.select(self.choice - 1)
            return False

        if e.key == K_DOWN:
            self.select(self.choice + 1)
            return False

        if e.key == K_RETURN:
            return True

        # Menus should be at the top of the focus list, except perhaps
        # the escape key (which triggers a new menu).
        return False

    def menu_layout(self):

        text = []

        if self.label:
            text.extend(_list_text(self.label, config.color_normal, None))

        for i, t in enumerate(self.choices):
            if i == self.choice:
                color = config.color_select
            else:
                color = config.color_choice

            text.extend(_list_text(t, color, i))

        return text
        
    def select(self, choice):

        if choice < 0:
            choice = 0

        if choice >= len(self.choices):
            choice = len(self.choices) - 1

        self.choice = choice

        self.set_text(self.menu_layout())

class LabelBox(TextBox):

    def __init__(self, text, width=None):    

        if width is None:
            width = config.label_box_width

        TextBox.__init__(self, width,
                         _list_text(text, config.color_normal, None))

class MessageBox(TextBox):

    def __init__(self, text, width=None, minheight=None):

        if width is None:
            width = config.message_box_width

        if minheight is None:
            minheight = config.message_box_minheight

        TextBox.__init__(self,
                         width,
                         _list_text(text, config.color_normal, None),
                         minheight=minheight)

    def mouse_event(self, button, x, y):
        if button:
            return True

    def key_event(self, e):

        if e.key == K_SPACE or e.key == K_RETURN:
            return True


class PromptBox(TextBox):
    def __init__(self, prompt, limit, default="",
                 width=None, minheight=None):

        if width is None:
            width = config.prompt_box_width

        if minheight is None:
            minheight = config.prompt_box_minheight

        self.prompt = prompt
        self.limit = limit
        self.content = default

        TextBox.__init__(self,
                         width,
                         self.layout_text(),
                         minheight=minheight)

    def layout_text(self):        

        rv = []

        rv.extend(_list_text(self.prompt, config.color_normal, None))

        rv.extend(_list_text(self.content + "_",
                             config.color_select,
                             None))
        return rv

    def key_event(self, e):

        c = str(e.unicode)

        if e.key == K_RETURN:
            return True

        if c.isalnum() or c.isspace():
            self.content += c

        if e.key == K_BACKSPACE:
            self.content = self.content[:-1]

        self.content = self.content[:self.limit]

        self.set_text(self.layout_text())

        return False

    def mouse_event(self, button, x, y):
        return False

def init():
    global font
    font = pygame.font.Font(config.font_file,
                            config.font_size)

    global font_linesize
    font_linesize = font.get_linesize()

    global font_height
    font_height = font.get_height()
    
    Box.class_init()
